
# start the measure
class SetClimateZoneFromWeatherData < OpenStudio::Measure::ModelMeasure
  # human readable name
  def name
    # Measure name should be the title case of the class name.
    return 'Set Climate Zone from Weather Data'
  end

  # human readable description
  def description
    return 'Determines the climate zone from the weather data file, following ASHRAE 169 steps.'
  end

  # human readable description of modeling approach
  def modeler_description
    return 'Based on a169_2006_climate_zone method of openstudio-standards library, this measure determines the climate zone computing heating and cooling design days from temperature data of the epw file instead of reading the corresponding STAT file.'
  end
  
  # define what happens when the measure is run
  def run(model, runner, user_arguments)
    super(model, runner, user_arguments)
    
    climate_zone = model.getClimateZones.getClimateZone("ASHRAE", 2006).value
    if climate_zone.empty? then
      runner.registerInitialCondition("No climate zone is set.")
    else
      runner.registerInitialCondition("The initial climate zone is #{climate_zone}.")
    end
    
    weather_file = model.getWeatherFile.file
    if weather_file.empty? then
      runner.registerError("No weather file is set.")
      return false
    end
    weather_file = weather_file.get
    
    data = weather_file.data
    if data.empty? then
      runner.registerError("Weather file doesn't have data.")
      return false
    end

    cdd10C = 0.0 # degreeDays
    hdd18C = 0.0 # degreeDays
    data.each do |epw_data_point|
      temperature = epw_data_point.dryBulbTemperature.get # degreeCelsius
      cdd10C += [temperature-10,0].max / 24 # degreeDays
      hdd18C += [18-temperature,0].max / 24 # degreeDays      
    end
    
    runner.registerInfo("CCD10ºC: #{cdd10C.round}")
    runner.registerInfo("HDD18ºC: #{hdd18C.round}")
    
    if 5000 < cdd10C
      climate_zone = "1A"
    elsif 3500 < cdd10C
      climate_zone = "2A"  
    elsif hdd18C < 2000
      if 2500 < cdd10C then
        climate_zone = "3C"
      else
        climate_zone = "3A"
      end
    elsif hdd18C < 3000
      if 1500 < cdd10C then
        climate_zone = "4C"
      else
        climate_zone = "4A"
      end
    elsif hdd18C < 4000
      if 1000 < cdd10C then
        climate_zone = "5C"
      else
        climate_zone = "5A"
      end
    elsif hdd18C < 5000
      climate_zone = "6A"
    elsif hdd18C < 7000
      climate_zone = "7"
    else
      climate_zone = "8"
    end
    
    runner.registerFinalCondition("The final climate zone is #{climate_zone}.")
    model.getClimateZones.setClimateZone("ASHRAE", 2006, climate_zone)
    return true
  end
end

# register the measure to be used by the application
SetClimateZoneFromWeatherData.new.registerWithApplication
